start_time_AQMS <- Sys.time()

# Loop Ordinary Kriging #######################################################

print("Loop Ordinary Kriging")

save_OK_result              <- paste0(save_result,"OK_monthly_",pollutant, "_",read_year)
model_raw_at_AQMS           <- matrix(NA,ncol=nrow(local_stations@data),nrow=0) #we will have a list with a data.frame for each hour
colnames(model_raw_at_AQMS) <- local_stations$ref_obs #matrix with empty data and the name of stations


# if set_model is FALSE, figure out which are the available_dates_model
if(set_model=="FALSE" & model_time_resolution=="daily"){
  available_dates_model <- daily_dates_format[!is.na(model[1,1,])]
}else if(set_model=="FALSE" & model_time_resolution=="hourly"){
  available_dates_model <- df_obs_sub$date[!is.na(model[1,1,])]
}

# Initialize snowfall
# Set the number of CPUs to use - adjust based on your system
sfInit(parallel = TRUE, cpus = 4)

# Export all necessary packages to the worker nodes
sfLibrary(sp)
sfLibrary(gstat)
sfLibrary(automap)
sfLibrary(plyr)
sfLibrary(dplyr)

# Main function to process each date
process_date <- function(i, all_dates, lat_model, lon_model, model, local_stations, 
                         crs_latlon, crs_utm, compute_in_UTM, variogram_ok, 
                         save_OK_result, time_resolution) {
  
  date_i <- all_dates[i]
  
  if(i != length(all_dates)) {
    date_ii <- all_dates[i+1]
  } else {
    date_ii <- date_i
  }
  
  print(paste("Processing date:", date_i))
  
  df_model <- model[,,i]
  
  # Initialize result container
  model_raw_at_AQMS <- NULL
  
  # Check if model data exists and is valid
  if (!any(is.na(df_model))) {
    
    # 1. Interpolate model values at obs locations
    # 1.1. Set spatialpointsdataframe
    df_model <- data.frame(lat = lat_model, lon = lon_model, pollutant = sapply(df_model, as.matrix))
    coordinates(df_model) <- ~lon+lat
    sp::proj4string(df_model) <- crs_latlon
    
    # Transform to UTM if needed
    if (compute_in_UTM) {
      df_model <- spTransform(df_model, sp::CRS(crs_utm))
    }
    
    # Set up observation points
    df_obs <- data.frame(lat = local_stations@coords[,2],
                         lon = local_stations@coords[,1],
                         pollutant = NA,
                         code = local_stations$ref_obs)
    coordinates(df_obs) <- ~lon+lat
    sp::proj4string(df_obs) <- crs_latlon
    
    # Transform to UTM if needed
    if (compute_in_UTM) {
      df_obs <- spTransform(df_obs, sp::CRS(crs_utm))
    }
    
    # 1.2 Compute Variogram
    set.seed(1) 
    pick_random_points <- sample(1:nrow(df_model), 1000)
    
    # Try to compute variogram
    pollutant.fit_ok <- tryCatch({
      automap::autofitVariogram(formula = pollutant~1, 
                                input_data = df_model[pick_random_points,],
                                fix.values = c(0, NA, NA),
                                miscFitOptions = list(min.np.bin = 1))
    }, error = function(e) {
      # If error, load saved variogram if it exists
      if (file.exists(variogram_ok)) {
        return(readRDS(variogram_ok))
      } else {
        stop("Variogram computation failed and no saved variogram exists")
      }
    })
    
    # Save variogram if it doesn't exist yet
    if (!file.exists(variogram_ok)) {
      saveRDS(pollutant.fit_ok, variogram_ok)
    }
    
    # 1.3 Compute ordinary Kriging
    ok_gstat <- gstat(formula = pollutant~1, data = df_model, model = pollutant.fit_ok$var_model, nmax = 100)
    
    # Interpolation of the model
    OK <- predict(object = ok_gstat, newdata = df_obs)
    
    # Force concentrations to be positive
    OK$var1.pred[OK$var1.pred < 0] <- 0
    
    # If variogram produces NaNs, use saved one
    if (any(is.na(OK@data$var1.pred))) {
      print("WARNING: VARIOGRAM DIVERGED! Using a precalculated VARIOGRAM")
      pollutant.fit_ok <- readRDS(variogram_ok)
      ok_gstat <- gstat(formula = pollutant~1, data = df_model, model = pollutant.fit_ok$var_model, nmax = 100)
      
      # Retry interpolation
      OK <- predict(object = ok_gstat, newdata = df_obs)
    }
    
    model_raw_at_AQMS_h <- matrix(data = OK$var1.pred, nrow = 1)
    colnames(model_raw_at_AQMS_h) <- df_obs$code
    
  } else {
    print(paste0("WARNING: The model is corrupted or non-existent in ", date_i))
    model_raw_at_AQMS_h <- matrix(data = NA, nrow = 1, ncol = nrow(local_stations@data))
    colnames(model_raw_at_AQMS_h) <- local_stations@data$ref_obs
  }
  
  # Determine month for saving
  if (time_resolution == "hourly") {
    date_i_posix <- as.POSIXct(date_i, format = "%Y%m%d %H:%M", tz = "UTC")
    date_ii_posix <- as.POSIXct(date_ii, format = "%Y%m%d %H:%M", tz = "UTC")
    month_num <- month(date_i_posix)
    next_month_num <- month(date_ii_posix)
  } else if (time_resolution == "daily") {
    date_i_date <- as.Date(date_i, format = "%Y%m%d", tz = "UTC")
    date_ii_date <- as.Date(date_ii, format = "%Y%m%d", tz = "UTC")
    month_num <- month(date_i_date)
    next_month_num <- month(date_ii_date)
  }
  
  # Return results for this date along with information needed for saving
  return(list(
    model_raw_at_AQMS_h = model_raw_at_AQMS_h,
    date_i = date_i,
    month_num = month_num,
    next_month_num = next_month_num
  ))
}

# Main execution
main <- function() {
  # Make sure output directory exists
  if (!file.exists(save_OK_result)) {
    dir.create(save_OK_result)
  }
  
  # Create index list for parallelization
  indices <- 1:length(all_dates)
  
  # Export all necessary variables to workers
  sfExport("all_dates", "lat_model", "lon_model", "model", "local_stations",
           "crs_latlon", "crs_utm", "compute_in_UTM", "variogram_ok", 
           "save_OK_result", "time_resolution")
  
  # Execute parallel processing
  results <- sfLapply(indices, function(i) {
    process_date(i, all_dates, lat_model, lon_model, model, local_stations,
                 crs_latlon, crs_utm, compute_in_UTM, variogram_ok,
                 save_OK_result, time_resolution)
  })
  
  # Process results and save monthly data
  current_month <- NULL
  model_raw_at_AQMS <- NULL
  
  for (i in 1:length(results)) {
    result <- results[[i]]
    
    # Initialize if first iteration or new month
    if (is.null(current_month) || current_month != result$month_num) {
      current_month <- result$month_num
      model_raw_at_AQMS <- result$model_raw_at_AQMS_h
    } else {
      # Add to existing month's data
      model_raw_at_AQMS <- plyr::rbind.fill(as.data.frame(model_raw_at_AQMS), 
                                            as.data.frame(result$model_raw_at_AQMS_h))
    }
    
    # Save if month changes or last iteration
    if (i == length(results) || result$month_num != result$next_month_num) {
      save(object = model_raw_at_AQMS, 
           file = paste0(save_OK_result, "/model_raw_at_AQMS_", "month_", current_month, ".rds"))
      
      # Reset for next month
      model_raw_at_AQMS <- NULL
    }
  }
  
  # Create final dataframe with dates
  dates_vector <- sapply(results, function(x) x$date_i)
  model_raw_values <- do.call(rbind, lapply(results, function(x) as.data.frame(x$model_raw_at_AQMS_h)))
  
  model_raw_at_AQMS_df <- cbind(date = dates_vector, model_raw_values)
  
  # Set timezone
  if (time_resolution == "hourly") {
    model_raw_at_AQMS_df$date <- as.POSIXct(model_raw_at_AQMS_df$date, format = "%Y%m%d %H:%M", tz = "UTC")
  } else {
    model_raw_at_AQMS_df$date <- as.Date(model_raw_at_AQMS_df$date, format = "%Y%m%d", tz = "UTC")
  }
  
  # Clean up parallel workers
  sfStop()
  
  return(model_raw_at_AQMS_df)
}

# Execute the main function
model_raw_at_AQMS_df <- main()

if (save_OK){
  saveRDS(object=model_raw_at_AQMS_df, file=paste0(save_result,"model_raw_at_AQMS_",file_format, "_",pollutant,"_df.rds"))
  print("model_raw_at_AQMS saved")
} else {
  print("model_raw_at_AQMS have not been saved")
}


end_time_AQMS    <- Sys.time()
temps_total_AQMS <- end_time_AQMS - start_time_AQMS
print(paste0("Interpolation of the model to AQMS, execution time:",temps_total_AQMS))

